@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2017 Fabian Groffen
+ * Copyright 2013-2018 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -21,7 +21,6 @@
 #include <string.h>
 #include <time.h>
 #include <math.h>
-#include <regex.h>
 #include <pthread.h>
 #include <errno.h>
 #include <assert.h>
@@ -32,6 +31,7 @@
 #include "router.h"
 #include "fnv1a.h"
 #include "aggregator.h"
+#include "posixregex.h"
 
 static pthread_t aggregatorid;
 static size_t prevreceived = 0;
@@ -261,10 +261,13 @@ aggregator_putmetric(
 				 * requested in issue #72.
 				 * For consistency with other tools/old carbon-aggregator
 				 * align the buckets to interval boundaries such that it is
-				 * predictable what intervals will be taken, issue #104. */
+				 * predictable what intervals will be taken, issue #104.
+				 * Reduce maximum splay to a couple of seconds at most,
+				 * issue #435. */
 				time(&now);
 				now = ((now - s->expire) / s->interval) * s->interval;
-				invocation->expire = s->expire + (rand() % s->interval);
+				invocation->expire = s->expire +
+					(rand() % s->interval < 10 ? s->interval : 10);
 
 				/* allocate enough buckets to hold the past + future */
 				invocation->buckets =
