@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2017 Fabian Groffen
+ * Copyright 2013-2022 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -29,7 +29,6 @@
 #include "collector.h"
 
 static dispatcher **dispatchers;
-static char debug = 0;
 static pthread_t collectorid;
 static char keep_running = 1;
 static router *pending_router = NULL;
@@ -44,6 +43,7 @@ collector_runner(void *s)
 	size_t totticks;
 	size_t totmetrics;
 	size_t totblackholes;
+	size_t totdiscards;
 	size_t totqueued;
 	size_t totstalls;
 	size_t totdropped;
@@ -51,6 +51,7 @@ collector_runner(void *s)
 	size_t ticks;
 	size_t metrics;
 	size_t blackholes;
+	size_t discards;
 	size_t queued;
 	size_t stalls;
 	size_t dropped;
@@ -75,16 +76,24 @@ collector_runner(void *s)
 	size_t (*d_ticks)(dispatcher *) = NULL;
 	size_t (*d_metrics)(dispatcher *) = NULL;
 	size_t (*d_blackholes)(dispatcher *) = NULL;
+	size_t (*d_discards)(dispatcher *) = NULL;
 	size_t (*d_sleeps)(dispatcher *) = NULL;
 	size_t (*a_received)(aggregator *) = NULL;
 	size_t (*a_sent)(aggregator *) = NULL;
 	size_t (*a_dropped)(aggregator *) = NULL;
 
 #define send(metric) \
-	if (debug & 1) \
+	if (mode & MODE_DEBUG) \
 		logout("%s", metric); \
-	else \
-		server_send(submission, strdup(metric), 1);
+	else { \
+		size_t len = strlen(metric); \
+		char *mtrc = malloc(sizeof(char) * len + sizeof(len)); \
+		if (mtrc != NULL) { \
+			*((size_t *)mtrc) = len; \
+			memcpy(mtrc + sizeof(len), metric, len); \
+			server_send(submission, mtrc, 1); \
+		} \
+	}
 
 	nextcycle = time(NULL) + collector_interval;
 	while (__sync_bool_compare_and_swap(&keep_running, 1, 1)) {
@@ -112,6 +121,7 @@ collector_runner(void *s)
 				d_ticks = dispatch_get_ticks_sub;
 				d_metrics = dispatch_get_metrics_sub;
 				d_blackholes = dispatch_get_blackholes_sub;
+				d_discards = dispatch_get_discards_sub;
 				d_sleeps = dispatch_get_sleeps_sub;
 				a_received = aggregator_get_received_sub;
 				a_sent = aggregator_get_sent_sub;
@@ -124,6 +134,7 @@ collector_runner(void *s)
 				d_ticks = dispatch_get_ticks;
 				d_metrics = dispatch_get_metrics;
 				d_blackholes = dispatch_get_blackholes;
+				d_discards = dispatch_get_discards;
 				d_sleeps = dispatch_get_sleeps;
 				a_received = aggregator_get_received;
 				a_sent = aggregator_get_sent;
@@ -149,18 +160,23 @@ collector_runner(void *s)
 		totticks = 0;
 		totmetrics = 0;
 		totblackholes = 0;
+		totdiscards = 0;
 		totsleeps = 0;
 		for (i = 0; dispatchers[i] != NULL; i++) {
 			totsleeps += sleeps = d_sleeps(dispatchers[i]);
 			totticks += ticks = d_ticks(dispatchers[i]);
 			totmetrics += metrics = d_metrics(dispatchers[i]);
 			totblackholes += blackholes = d_blackholes(dispatchers[i]);
+			totdiscards += discards = d_discards(dispatchers[i]);
 			snprintf(m, sizem, "dispatcher%d.metricsReceived %zu %zu\n",
 					i + 1, metrics, (size_t)now);
 			send(metric);
 			snprintf(m, sizem, "dispatcher%d.metricsBlackholed %zu %zu\n",
 					i + 1, blackholes, (size_t)now);
 			send(metric);
+			snprintf(m, sizem, "dispatcher%d.metricsDiscarded %zu %zu\n",
+					i + 1, discards, (size_t)now);
+			send(metric);
 			snprintf(m, sizem, "dispatcher%d.wallTime_us %zu %zu\n",
 					i + 1, ticks, (size_t)now);
 			send(metric);
@@ -174,6 +190,9 @@ collector_runner(void *s)
 		snprintf(m, sizem, "metricsBlackholed %zu %zu\n",
 				totblackholes, (size_t)now);
 		send(metric);
+		snprintf(m, sizem, "metricsDiscarded %zu %zu\n",
+				totdiscards, (size_t)now);
+		send(metric);
 		snprintf(m, sizem, "dispatch_wallTime_us %zu %zu\n",
 				totticks, (size_t)now);
 		send(metric);
@@ -279,7 +298,7 @@ collector_runner(void *s)
 			send(metric);
 		}
 
-		if (debug & 1)
+		if (mode & MODE_DEBUG)
 			fflush(stdout);
 	}
 
@@ -323,7 +342,7 @@ collector_writer(void *unused)
 		}
 		assert(srvs != NULL);
 		sleep(1);
-		if (debug & 1) {
+		if (mode & MODE_METRICSTAT) {
 			size_t mpsout;
 			size_t totout;
 			size_t mpsdrop;
@@ -385,6 +404,7 @@ collector_writer(void *unused)
 					totdisc - lastdisc, totdisc,
 					(int)(((double)dticks * 100.0) / (double)(dticks + dsleeps))
 				  );
+			fflush(stdout);
 			lastconn = totconn;
 			lastdisc = totdisc;
 		}
@@ -448,20 +468,17 @@ collector_reloadcomplete(void)
  * Initialises and starts the collector.
  */
 void
-collector_start(dispatcher **d, router *rtr, server *submission, char cum)
+collector_start(dispatcher **d, router *rtr, server *submission)
 {
 	dispatchers = d;
 	collector_schedulereload(rtr);
 
-	if (mode & MODE_DEBUG)
-		debug = 1;
-	debug |= (cum ? 0 : 2);
-
 	if (mode & MODE_SUBMISSION) {
 		if (pthread_create(&collectorid, NULL, collector_writer, NULL) != 0)
 			logerr("failed to start collector!\n");
 	} else {
-		if (pthread_create(&collectorid, NULL, collector_runner, submission) != 0)
+		if (pthread_create(&collectorid, NULL,
+						   collector_runner, submission) != 0)
 			logerr("failed to start collector!\n");
 	}
 }
