@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2017 Fabian Groffen
+ * Copyright 2013-2022 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -27,7 +27,7 @@
 #define timediff(X, Y) \
 	(Y.tv_sec > X.tv_sec ? (Y.tv_sec - X.tv_sec) * 1000 * 1000 + ((Y.tv_usec - X.tv_usec)) : Y.tv_usec - X.tv_usec)
 
-void collector_start(dispatcher **d, router *rtr, server *submission, char cum);
+void collector_start(dispatcher **d, router *rtr, server *submission);
 void collector_stop(void);
 void collector_schedulereload(router *rtr);
 char collector_reloadcomplete(void);
