@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2021 Fabian Groffen
+ * Copyright 2013-2024 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -124,10 +124,10 @@ typedef struct _connection {
 	char srcaddr[24];  /* string representation of source address */
 	char buf[METRIC_BUFSIZ];
 	int buflen;
-	char needmore:1;
-	char noexpire:1;
-	char isaggr:1;
-	char isudp:1;
+	unsigned char needmore:1;
+	unsigned char noexpire:1;
+	unsigned char isaggr:1;
+	unsigned char isudp:1;
 	char datawaiting; /* full byte for atomic access */
 	char metric[METRIC_BUFSIZ];
 	destination dests[CONN_DESTS_SIZE];
@@ -409,11 +409,9 @@ lzreadbuf(z_strm *strm, void *buf, size_t sze, int rval, int err)
 		return -1;
 	}
 
-	/* if we decompressed something, update our ibuf */
+	strm->hdl.lz4.iloc += srcsize;
 
-	if (destsize > 0) {
-		strm->hdl.lz4.iloc += srcsize;
-	} else if (destsize == 0) {
+	if (destsize == 0) {
 		tracef("No LZ4 data was produced\n");
 		errno = err ? err : EAGAIN;
 		return -1;
@@ -643,7 +641,7 @@ dispatch_transplantlistener(listener *olsnr, listener *nlsnr, router *r)
 		if (listeners[c] == olsnr) {
 			router_transplant_listener_socks(r, olsnr, nlsnr);
 #ifdef HAVE_SSL
-			if ((nlsnr->transport & ~0xFFFF) == W_SSL)
+			if (nlsnr->transport & W_SSL)
 				nlsnr->ctx = olsnr->ctx;
 #endif
 			if (olsnr->saddrs) {
@@ -657,8 +655,6 @@ dispatch_transplantlistener(listener *olsnr, listener *nlsnr, router *r)
 	pthread_rwlock_unlock(&listenerslock);
 }
 
-#define CONNGROWSZ  1024
-
 /**
  * Adds a connection socket to the chain of connections.
  * Connection sockets are those which need to be read from.
@@ -674,16 +670,19 @@ dispatch_addconnection(int sock, listener *lsnr)
 	int compress_type;
 	char *ibuf;
 #endif
+	char checksize;
 
 	pthread_rwlock_rdlock(&connectionslock);
 	for (c = 0; c < connectionslen; c++)
 		if (__sync_bool_compare_and_swap(&(connections[c].takenby),
 					C_FREE, C_SETUP))
 			break;
+	checksize = c == connectionslen;
 	pthread_rwlock_unlock(&connectionslock);
 
-	if (c == connectionslen) {
+	if (checksize) {
 		connection *newlst;
+		size_t      growlen;
 
 		pthread_rwlock_wrlock(&connectionslock);
 		if (connectionslen > c) {
@@ -691,8 +690,17 @@ dispatch_addconnection(int sock, listener *lsnr)
 			pthread_rwlock_unlock(&connectionslock);
 			return dispatch_addconnection(sock, lsnr);
 		}
+		/* take it slow with extending connections, because each
+		 * connection struct is 65K, so use an exponential approach
+		 * ceiled by CONNGROWSZ */
+		growlen = connectionslen * 5 / 10;
+		c = mode & MODE_SUBMISSION ? 2 : 10;
+		if (growlen < c)
+			growlen = c;
+		else if (growlen > CONNGROWSZ)
+			growlen = CONNGROWSZ;
 		newlst = realloc(connections,
-				sizeof(connection) * (connectionslen + CONNGROWSZ));
+				sizeof(connection) * (connectionslen + growlen));
 		if (newlst == NULL) {
 			logerr("cannot add new connection: "
 					"out of memory allocating more slots (max = %zu)\n",
@@ -711,14 +719,14 @@ dispatch_addconnection(int sock, listener *lsnr)
 			}
 		}
 
-		for (c = connectionslen; c < connectionslen + CONNGROWSZ; c++) {
+		for (c = connectionslen; c < connectionslen + growlen; c++) {
 			memset(&newlst[c], '\0', sizeof(connection));
 			newlst[c].takenby = C_FREE;
 		}
 		connections = newlst;
 		c = connectionslen;  /* for the setup code below */
 		newlst[c].takenby = C_SETUP;
-		connectionslen += CONNGROWSZ;
+		connectionslen += growlen;
 
 		pthread_rwlock_unlock(&connectionslock);
 	}
@@ -759,7 +767,7 @@ dispatch_addconnection(int sock, listener *lsnr)
 	/* set socket or SSL connection */
 	connections[c].strm->nextstrm = NULL;
 	connections[c].strm->strmreadbuf = NULL;
-	if (lsnr == NULL || (lsnr->transport & ~0xFFFF) != W_SSL) {
+	if (lsnr == NULL || !(lsnr->transport & W_SSL)) {
 		if (lsnr == NULL || lsnr->ctype != CON_UDP) {
 			connections[c].strm->hdl.sock = sock;
 			connections[c].strm->strmread = &sockread;
@@ -785,6 +793,13 @@ dispatch_addconnection(int sock, listener *lsnr)
 		}
 		SSL_set_fd(connections[c].strm->hdl.ssl, sock);
 		SSL_set_accept_state(connections[c].strm->hdl.ssl);
+		if (lsnr->transport & W_MTLS) {  /* issue #444 */
+			SSL_set_verify(connections[c].strm->hdl.ssl,
+						   SSL_VERIFY_PEER |
+						   SSL_VERIFY_FAIL_IF_NO_PEER_CERT |
+						   SSL_VERIFY_CLIENT_ONCE,
+						   NULL);
+		}
 
 		connections[c].strm->strmread = &sslread;
 		connections[c].strm->strmclose = &sslclose;
@@ -1020,7 +1035,7 @@ dispatch_received_metrics(connection *conn, dispatcher *self)
 	q = conn->metric;
 	firstspace = NULL;
 	lastnl = NULL;
-	search_tags = self->tags_supported;
+	search_tags = self->tags_supported ? 1 : 0;
 	for (p = conn->buf; p - conn->buf < conn->buflen; p++) {
 		if (*p == '\n' || *p == '\r') {
 			/* end of metric */
@@ -1060,14 +1075,16 @@ dispatch_received_metrics(connection *conn, dispatcher *self)
 			/* restart building new one from the start */
 			q = conn->metric;
 			firstspace = NULL;
-			search_tags = self->tags_supported;
+			search_tags = self->tags_supported ? 1 : 0;
 
 			gettimeofday(&conn->lastwork, NULL);
 			conn->maxsenddelay = 0;
 			/* send the metric to where it is supposed to go */
 			if (dispatch_process_dests(conn, self, conn->lastwork) == 0)
 				break;
-		} else if (*p == ' ' || *p == '\t' || *p == '.') {
+		} else if (search_tags != 2 && /* leave tags alone, issue #453 */
+				   (*p == ' ' || *p == '\t' || *p == '.'))
+		{
 			/* separator */
 			if (q == conn->metric) {
 				/* make sure we skip this on next iteration to
@@ -1075,6 +1092,7 @@ dispatch_received_metrics(connection *conn, dispatcher *self)
 				lastnl = p;
 				continue;
 			}
+
 			if (*p == '\t')
 				*p = ' ';
 			if (*p == ' ' && firstspace == NULL) {
@@ -1089,9 +1107,9 @@ dispatch_received_metrics(connection *conn, dispatcher *self)
 				if (*(q - 1) != *p && (q - 1) != firstspace)
 					*q++ = *p;
 			}
-		} else if (search_tags && *p == ';') {
+		} else if (search_tags == 1 && *p == ';') {
 			/* copy up to next space */
-			search_tags = 0;
+			search_tags = 2;
 			firstspace = q;
 			*q++ = *p;
 		} else if (
