@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2022 Fabian Groffen
+ * Copyright 2013-2024 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -64,7 +64,6 @@ static int optimiserthreshold = 50;
 static int sockbufsize = 0;
 static int collector_interval = 60;
 static unsigned int listenbacklog = 32;
-static col_mode smode = CUM;
 static dispatcher **workers = NULL;
 static char workercnt = 0;
 static router *rtr = NULL;
@@ -189,6 +188,8 @@ do_reload(void)
 		/* no changes, shortcut */
 		logout("no config changes found\n");
 		router_free(newrtr);
+		/* check if there are file targets, make them reopen files #342 */
+		router_reopen_files(rtr);
 		logout("SIGHUP handler complete\n");
 		return;
 	}
@@ -402,7 +403,8 @@ do_usage(char *name, int exitcode)
 	printf("  -q  server queue size, defaults to %d\n", queuesize);
 	printf("  -L  server max stalls, defaults to %d\n", maxstalls);
 #ifdef HAVE_SSL
-	printf("  -C  use CA <cert> to verify SSL connections\n");
+	printf("  -C  use CA <cert> to verify outgoing SSL connections or\n");
+	printf("      incoming mTLS connections\n");
 #endif
 	printf("  -B  connection listen backlog, defaults to 32\n");
 	printf("  -U  socket receive buffer size, max/min/default values depend on OS\n");
@@ -419,11 +421,13 @@ do_usage(char *name, int exitcode)
 #endif
 	printf("  -s  submission mode: don't add any metrics to the stream like\n"
 	       "      statistics, report drop counts and queue pressure to log\n");
+	printf("  -S  implies submission mode, print iostat-like statistics\n");
 	printf("  -t  config test mode: prints rule matches from input on stdin\n");
 	printf("  -H  hostname: override hostname (used in statistics)\n");
 	printf("  -D  daemonise: detach and run in background\n");
 	printf("  -P  pidfile: write a pid to a specified pidfile\n");
-	printf("  -O  minimum number of rules before optimising the ruleset, default: %d\n", optimiserthreshold); 
+	printf("  -O  minimum number of rules before optimising the ruleset, "
+		   "default: %d\n", optimiserthreshold); 
 
 	exit(exitcode);
 }
@@ -448,7 +452,7 @@ main(int argc, char * const argv[])
 		snprintf(relay_hostname, sizeof(relay_hostname), "127.0.0.1");
 
 	while ((ch = getopt(argc, argv,
-					":hvdstf:l:p:w:b:q:L:C:T:c:m:M:H:B:U:EDP:O:")) != -1)
+					":hvdsStf:l:p:w:b:q:L:C:T:c:m:M:H:B:U:EDP:O:")) != -1)
 	{
 		switch (ch) {
 			case 'v':
@@ -467,6 +471,9 @@ main(int argc, char * const argv[])
 			case 's':
 				mode |= MODE_SUBMISSION;
 				break;
+			case 'S':
+				mode |= MODE_SUBMISSION | MODE_METRICSTAT;
+				break;
 			case 't':
 				/* -t: test interactively, -tt: test config and exit */
 				if (mode & MODE_TEST) {
@@ -992,8 +999,8 @@ main(int argc, char * const argv[])
 	/* server used for delivering metrics produced inside the relay,
 	 * that is, the collector (statistics) */
 	if ((internal_submission = server_new(
-					"internal", listenport, T_LINEMODE, W_PLAIN, CON_PIPE,
-					NULL, NULL, 3000,
+					"internal", listenport, T_LINEMODE, W_PLAIN,
+					NULL, NULL, CON_PIPE, NULL, NULL, 3000,
 					batchsize, maxstalls, iotimeout, sockbufsize)) == NULL)
 	{
 		logerr("failed to create internal submission queue, shutting down\n");
@@ -1016,7 +1023,7 @@ main(int argc, char * const argv[])
 
 	logout("starting statistics collector\n");
 	if (internal_submission != NULL)
-		collector_start(&workers[1], rtr, internal_submission, smode == CUM);
+		collector_start(&workers[1], rtr, internal_submission);
 
 	logout("starting servers\n");
 	if (router_start(rtr) != 0) {
