@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2022 Fabian Groffen
+ * Copyright 2013-2024 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -397,7 +397,7 @@ char *
 router_validate_address(
 		router *rtr,
 		char **retip, unsigned short *retport, void **retsaddr,
-		void **rethint, char *ip, con_proto proto)
+		void **rethint, char *ip, con_proto proto, char allowanyaddr)
 {
 	unsigned short port = GRAPHITE_PORT;
 	struct addrinfo *saddr = NULL;
@@ -435,8 +435,13 @@ router_validate_address(
 			return(ra_strdup(rtr->a, "expected ']'"));
 		}
 	}
-	if (lastcolon == ip)
-		ip = NULL;  /* only resolve port, e.g. any interface */
+	if (lastcolon == ip) {
+		if (allowanyaddr) {
+			ip = NULL;  /* only resolve port, e.g. any interface */
+		} else {
+			return(ra_strdup(rtr->a, "host or IP-address required"));
+		}
+	}
 
 	memset(&hint, 0, sizeof(hint));
 	saddr = NULL;
@@ -454,20 +459,14 @@ router_validate_address(
 		int err;
 		/* now resolve this the normal way to check validity */
 		hint.ai_flags = AI_NUMERICSERV;
-		if ((err = getaddrinfo(ip, sport, &hint, &saddr)) != 0) {
-			char errmsg[512];
-			snprintf(errmsg, sizeof(errmsg), "warning: failed to resolve "
-					"%s, port %s, proto %s: %s",
-					ip == NULL ? "<any>" : ip, sport,
-					proto == CON_UDP ? "udp" : "tcp",
-					gai_strerror(err));
+		if ((err = getaddrinfo(ip, sport, &hint, &saddr)) != 0)
 			saddr = NULL;
-		}
 
 		/* no ra_malloc, it gets owned by server */
 		*rethint = malloc(sizeof(hint));
 		if (*rethint == NULL) {
-			freeaddrinfo(saddr);
+			if (saddr != NULL)
+				freeaddrinfo(saddr);
 			return(ra_strdup(rtr->a, "out of memory copying hint structure"));
 		}
 		memcpy(*rethint, &hint, sizeof(hint));
@@ -590,6 +589,8 @@ router_add_server(
 		char *inst,
 		con_type type,
 		con_trnsp transport,
+		char *mtlspemcert,
+		char *mtlspemkey,
 		con_proto proto,
 		struct addrinfo *saddrs,
 		struct addrinfo *hint,
@@ -608,7 +609,12 @@ router_add_server(
 		servers *s;
 		lwalk = NULL;
 
-		if (useall) {
+		/* Since #293 unresolvable targets are allowed, but when there's
+		 * only one target such as in #455 we end up here crashing
+		 * because walk is NULL.  So, in that scenario, basically ignore
+		 * the useall, and if the address lateron becomes resolvable,
+		 * treat it as single target. */
+		if (useall && walk != NULL) {
 			/* serialise the IP address, to make the targets explicit
 			 * (since we're expanding all A/AAAA records) */
 			saddr_ntop(walk, hnbuf);
@@ -633,7 +639,8 @@ router_add_server(
 		}
 		if (newserver == NULL) {
 			newserver = server_new(ip, port,
-					type, transport, proto, walk, hint,
+					type, transport, mtlspemcert, mtlspemkey,
+					proto, walk, hint,
 					ret->conf.queuesize, ret->conf.batchsize,
 					ret->conf.maxstalls, ret->conf.iotimeout,
 					ret->conf.sockbufsize);
@@ -1392,7 +1399,7 @@ router_readconfig(router *orig,
 			hint = NULL;
 			snprintf(sockbuf, sizeof(sockbuf), ":%u", listenport);
 			router_validate_address(ret, &ip, &port, &saddrs, &hint,
-					sockbuf, CON_TCP);
+					sockbuf, CON_TCP, 1 /* use any address */);
 			free(hint);
 			router_add_listener(ret, T_LINEMODE, W_PLAIN, NULL, 0, 0,
 								NULL, NULL, CON_TCP, ip, port, saddrs);
@@ -1400,7 +1407,7 @@ router_readconfig(router *orig,
 			hint = NULL;
 			snprintf(sockbuf, sizeof(sockbuf), ":%u", listenport);
 			router_validate_address(ret, &ip, &port, &saddrs, &hint,
-					sockbuf, CON_UDP);
+					sockbuf, CON_UDP, 1 /* use any address */);
 			free(hint);
 			router_add_listener(ret, T_LINEMODE, W_PLAIN, NULL, 0, 0,
 								NULL, NULL, CON_UDP, ip, port, saddrs);
@@ -1937,7 +1944,8 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 	server_transport(s->server) == W_PLAIN ? "" : " transport", \
 	(server_transport(s->server) & 0xFFFF) == W_PLAIN ? "" : \
 		con_trnsp_str[server_transport(s->server) & 0xFFFF], \
-	(server_transport(s->server) & ~0xFFFF) == W_SSL ? " ssl" : ""
+	(server_transport(s->server) & W_MTLS) ? " mtls" : \
+	(server_transport(s->server) & W_SSL) ? " ssl" : ""
 
 	if (rtr->listeners != NULL) {
 		listener *walk;
@@ -1949,8 +1957,10 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 					fprintf(f, " transport %s",
 							con_trnsp_str[walk->transport & 0xFFFF]);
 #ifdef HAVE_SSL
-				if ((walk->transport & ~0xFFFF) == W_SSL) {
-					fprintf(f, " ssl %s", router_quoteident(walk->pemcert));
+				if (walk->transport & W_SSL) {
+					fprintf(f, " %s %s",
+							walk->transport & W_MTLS ? "mtls" : "ssl",
+							router_quoteident(walk->pemcert));
 					if (walk->protomin != _rp_UNSET)
 						fprintf(f, " protomin %s",
 								router_tlsprotostr(walk->protomin));
@@ -2338,6 +2348,22 @@ router_transplant_queues(router *new, router *old)
 	}
 }
 
+/**
+ * Signal all file target servers to close their files.  Doing so means
+ * on the next event their file is re-opened.  This allows logrotate to
+ * move the file and HUP to start a new one.  #342
+ */
+void
+router_reopen_files(router *rtr)
+{
+	servers *s;
+
+	for (s = rtr->srvrs; s != NULL; s = s->next) {
+		if (server_ctype(s->server) == CON_FILE)
+			server_closecon(s->server);
+	}
+}
+
 inline listener *
 router_get_listeners(router *rtr)
 {
@@ -2409,7 +2435,7 @@ router_contains_listener(router *rtr, listener *lsnr)
 				if (lsnr->port == rwalk->port && strcmp(l, r) == 0) {
 #ifdef HAVE_SSL
 					/* check pemmtimespec */
-					if ((lsnr->transport & ~0xFFFF) == W_SSL &&
+					if (lsnr->transport & W_SSL &&
 							(lsnr->pemmtimespec.tv_sec !=
 									rwalk->pemmtimespec.tv_sec || 
 							lsnr->pemmtimespec.tv_nsec !=
@@ -2888,9 +2914,12 @@ router_route_intern(
 					}	break;
 					case ANYOF: {
 						/* we queue the same metrics at the same server */
-						unsigned int hash;
+						server         *s;
+						unsigned int    hash;
+						unsigned int    pos;
+						unsigned short  i;
 						
-						fnv1a_32(hash, p, metric, firstspace);
+						failif(retsize, *curlen + 1);
 
 						/* We could use the retry approach here, but since
 						 * our c is very small compared to MAX_INT, the bias
@@ -2898,10 +2927,27 @@ router_route_intern(
 						 * (MAX_INT % c) can be considered neglicible given
 						 * the number of occurances of c in the range of
 						 * MAX_INT, therefore we stick with a simple mod. */
-						hash %= d->cl->members.anyof->count;
-						failif(retsize, *curlen + 1);
-						ret[*curlen].dest =
-							d->cl->members.anyof->servers[hash];
+						fnv1a_32(hash, p, metric, firstspace);
+
+						/* find first non-failed server from here, but
+						 * be careful not to dump everything onto the
+						 * next neighbour */
+						pos   = hash % d->cl->members.anyof->count;
+						hash += rand();
+						for (i = 0; i < d->cl->members.anyof->count; i++) {
+							s = d->cl->members.anyof->servers[pos];
+							if (!server_failed(s)) {
+								ret[*curlen].dest = s;
+								break;
+							}
+							pos = (hash + i + 1) % d->cl->members.anyof->count;
+						}
+						if (ret[*curlen].dest == NULL) {
+							/* all failed, take original matching server */
+							ret[*curlen].dest =
+								d->cl->members.anyof->servers[hash];
+						}
+
 						produce_metric(ret[*curlen]);
 						set_metric(ret[*curlen]);
 						(*curlen)++;
@@ -2909,21 +2955,23 @@ router_route_intern(
 					}	break;
 					case FAILOVER: {
 						/* queue at the first non-failing server */
-						unsigned short i;
+						server         *s;
+						unsigned short  i;
 
 						failif(retsize, *curlen + 1);
 						ret[*curlen].dest = NULL;
 						for (i = 0; i < d->cl->members.anyof->count; i++) {
-							server *s = d->cl->members.anyof->servers[i];
+							s = d->cl->members.anyof->servers[i];
 							if (server_failed(s))
 								continue;
 							ret[*curlen].dest = s;
 							break;
 						}
-						if (ret[*curlen].dest == NULL)
+						if (ret[*curlen].dest == NULL) {
 							/* all failed, take first server */
 							ret[*curlen].dest =
 								d->cl->members.anyof->servers[0];
+						}
 
 						produce_metric(ret[*curlen]);
 						set_metric(ret[*curlen]);
