@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2017 Fabian Groffen
+ * Copyright 2013-2022 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -44,28 +44,58 @@
 #endif
 #ifdef HAVE_LZ4
 #include <lz4.h>
+#include <lz4frame.h>
+#endif
+#ifdef HAVE_SNAPPY
+#include <snappy-c.h>
 #endif
 #ifdef HAVE_SSL
 #include <openssl/ssl.h>
 #include <openssl/err.h>
 #endif
 
+#define FAIL_WAIT_TIME          6  /* 6 * 250ms = 1.5s */
+#define DISCONNECT_WAIT_TIME   12  /* 12 * 250ms = 3s */
+#define LEN_CRITICAL(Q)        (queue_free(Q) < self->bsize)
+
+typedef struct _z_strm {
+	ssize_t (*strmwrite)(struct _z_strm *, const void *, size_t);
+	int (*strmflush)(struct _z_strm *);
+	int (*strmclose)(struct _z_strm *);
+	const char *(*strmerror)(struct _z_strm *, int);     /* get last err str */
+	struct _z_strm *nextstrm;                            /* set when chained */
+#if defined(HAVE_LZ4) || defined(HAVE_SNAPPY) || defined(HAVE_GZIP)
+	char obuf[METRIC_BUFSIZ];
+	int obuflen;
+#endif
+#ifdef HAVE_SSL
+	SSL_CTX *ctx;
+#endif
+	union {
+#ifdef HAVE_GZIP
+		z_streamp gz;
+#endif
+#ifdef HAVE_LZ4
+		struct {
+			void *cbuf;
+			size_t cbuflen;
+		} z;
+#endif
+#ifdef HAVE_SSL
+		SSL *ssl;
+#endif
+		int sock;
+	} hdl;
+} z_strm;
+
 struct _server {
 	const char *ip;
 	unsigned short port;
 	char *instance;
 	struct addrinfo *saddr;
 	struct addrinfo *hint;
-	char reresolve:1;
 	int fd;
-	void *strm;
-	ssize_t (*strmwrite)(void *, const void *, size_t);  /* write impl */
-	int (*strmflush)(void *);                            /* flush impl */
-	int (*strmclose)(void *);                            /* close impl */
-	const char *(*strmerror)(void *, int);               /* get last err str */
-#ifdef HAVE_SSL
-	SSL_CTX *ctx;
-#endif
+	z_strm *strm;
 	queue *queue;
 	size_t bsize;
 	short iotimeout;
@@ -74,14 +104,18 @@ struct _server {
 	const char **batch;
 	con_type type;
 	con_trnsp transport;
+	char *mtlspemcert;
+	char *mtlspemkey;
 	con_proto ctype;
 	pthread_t tid;
 	struct _server **secondaries;
 	size_t secondariescnt;
-	char failover:1;
+	unsigned char reresolve:1;
+	unsigned char failover:1;
 	char failure;       /* full byte for atomic access */
 	char running;       /* full byte for atomic access */
 	char keep_running;  /* full byte for atomic access */
+	char reopen_con;    /* full byte for atomic access */
 	unsigned char stallseq;  /* full byte for atomic access */
 	size_t metrics;
 	size_t dropped;
@@ -96,39 +130,29 @@ struct _server {
 
 /* connection specific writers and closers */
 
-typedef struct _z_strm {
-	char obuf[METRIC_BUFSIZ];
-	int sock;
-	union {
-#ifdef HAVE_LZ4
-		LZ4_stream_t *lz;
-#endif
-		void *dummy;
-	} hdl;
-} z_strm;
-
 /* ordinary socket */
 static inline ssize_t
-sockwrite(void *strm, const void *buf, size_t sze)
+sockwrite(z_strm *strm, const void *buf, size_t sze)
 {
-	return write(*((int *)strm), buf, sze);
+	return write(strm->hdl.sock, buf, sze);
 }
 
 static inline int
-sockflush(void *strm)
+sockflush(z_strm *strm)
 {
 	/* noop, we don't use a stream in the normal case */
+	(void)strm;
 	return 0;
 }
 
 static inline int
-sockclose(void *strm)
+sockclose(z_strm *strm)
 {
-	return close(*((int *)strm));
+	return close(strm->hdl.sock);
 }
 
 static inline const char *
-sockerror(void *strm, int rval)
+sockerror(z_strm *strm, int rval)
 {
 	(void)strm;
 	(void)rval;
@@ -137,94 +161,319 @@ sockerror(void *strm, int rval)
 
 #ifdef HAVE_GZIP
 /* gzip wrapped socket */
+static inline int gzipflush(z_strm *strm);
+
 static inline ssize_t
-gzipwrite(void *strm, const void *buf, size_t sze)
+gzipwrite(z_strm *strm, const void *buf, size_t sze)
 {
-	return (ssize_t)gzwrite((gzFile)strm, buf, (unsigned)sze);
+	/* ensure we have space available */
+	if (strm->obuflen + sze > METRIC_BUFSIZ)
+		if (gzipflush(strm) != 0)
+			return -1;
+
+	/* append metric to buf */
+	memcpy(strm->obuf + strm->obuflen, buf, sze);
+	strm->obuflen += sze;
+
+	return sze;
 }
 
 static inline int
-gzipflush(void *strm)
+gzipflush(z_strm *strm)
 {
-	return gzflush((gzFile)strm, Z_SYNC_FLUSH);
+	int cret;
+	char cbuf[METRIC_BUFSIZ];
+	size_t cbuflen;
+	char *cbufp;
+	int oret;
+
+	strm->hdl.gz->next_in = (Bytef *)strm->obuf;
+	strm->hdl.gz->avail_in = strm->obuflen;
+	strm->hdl.gz->next_out = (Bytef *)cbuf;
+	strm->hdl.gz->avail_out = sizeof(cbuf);
+
+	do {
+		cret = deflate(strm->hdl.gz, Z_PARTIAL_FLUSH);
+		if (cret == Z_OK && strm->hdl.gz->avail_out == 0) {
+			/* too large output block, unlikely given the input, discard */
+			/* TODO */
+			break;
+		}
+		if (cret == Z_OK) {
+			cbufp = cbuf;
+			cbuflen = sizeof(cbuf) - strm->hdl.gz->avail_out;
+			while (cbuflen > 0) {
+				oret = strm->nextstrm->strmwrite(strm->nextstrm,
+						cbufp, cbuflen);
+				if (oret < 0)
+					return -1;  /* failure is failure */
+
+				/* update counters to possibly retry the remaining bit */
+				cbufp += oret;
+				cbuflen -= oret;
+			}
+
+			if (strm->hdl.gz->avail_in == 0)
+				break;
+
+			strm->hdl.gz->next_out = (Bytef *)cbuf;
+			strm->hdl.gz->avail_out = sizeof(cbuf);
+		} else {
+			/* discard */
+			break;
+		}
+	} while (1);
+	/* flush whatever we wrote */
+	strm->nextstrm->strmflush(strm->nextstrm);
+
+	/* reset the write position, from this point it will always need to
+	 * restart */
+	strm->obuflen = 0;
+
+	if (cret != Z_OK)
+		return -1;  /* we must reset/free gzip */
+
+	return 0;
 }
 
 static inline int
-gzipclose(void *strm)
+gzipclose(z_strm *strm)
 {
-	return gzclose((gzFile)strm);
+	int ret = strm->nextstrm->strmclose(strm->nextstrm);
+	deflateEnd(strm->hdl.gz);
+	free(strm->hdl.gz);
+	return ret;
+}
+
+static inline const char *
+gziperror(z_strm *strm, int rval)
+{
+	return strm->nextstrm->strmerror(strm->nextstrm, rval);
 }
 #endif
 
 #ifdef HAVE_LZ4
 /* lz4 wrapped socket */
-static inline ssize_t
-lzwrite(void *strm, const void *buf, size_t sze)
-{
-	int oret;
-	char *obuf = ((z_strm *)strm)->obuf;
-	int cret;
 
-	cret = LZ4_compress_fast_continue(((z_strm *)strm)->hdl.lz,
-			buf, obuf, sze, METRIC_BUFSIZ, 0);
-	if (cret == 0)
-		return -1;  /* we must reset/free lz */
-	while (cret > 0) {
-		oret = write(((z_strm *)strm)->sock, obuf, cret);
-		if (oret < 0)
-			return -1;  /* failure is failure */
-		/* update counters to possibly retry the remaining bit */
-		obuf += oret;
-		cret -= oret;
+static inline int lzflush(z_strm *strm);
+
+static inline ssize_t
+lzwrite(z_strm *strm, const void *buf, size_t sze)
+{
+	size_t towrite = sze;
+
+	/* use the same strategy as gzip: fill the buffer until space
+	   runs out. we completely fill the output buffer before flushing */
+
+	while (towrite > 0) {
+
+		size_t avail = METRIC_BUFSIZ - strm->obuflen;
+		size_t copysize = towrite > avail ? avail : towrite;
+
+		/* copy into the output buffer as much as we can */
+
+		if (copysize > 0) {
+			memcpy(strm->obuf + strm->obuflen, buf, copysize);
+			strm->obuflen += copysize;
+			towrite -= copysize;
+			buf += copysize;
+		}
+
+		/* if output buffer is full & still have bytes to write, flush now */
+
+		if (strm->obuflen == METRIC_BUFSIZ && towrite > 0 && lzflush(strm) != 0) {
+			logerr("Failed to flush LZ4 data to make space\n");
+			return -1;
+		}
 	}
 
 	return sze;
 }
 
 static inline int
-lzflush(void *strm)
+lzflush(z_strm *strm)
 {
+	int oret;
+	size_t ret;
+
+	/* anything to do? */
+
+	if (strm->obuflen == 0)
+		return 0;
+
+	/* the buffered data goes out as a single frame */
+
+	ret = LZ4F_compressFrame(strm->hdl.z.cbuf, strm->hdl.z.cbuflen, strm->obuf, strm->obuflen, NULL);
+	if (LZ4F_isError(ret)) {
+		logerr("Failed to compress %d LZ4 bytes into a frame: %d\n",
+		       strm->obuflen, (int)ret);
+		return -1;
+	}
+		/* write and flush */
+
+	if ((oret = strm->nextstrm->strmwrite(strm->nextstrm,
+					      strm->hdl.z.cbuf, ret)) < 0) {
+		logerr("Failed to write %lu bytes of compressed LZ4 data: %d\n",
+		       ret, oret);
+		return oret;
+	}
+
+	strm->nextstrm->strmflush(strm->nextstrm);
+
+	/* the buffer is gone. reset the write position */
+
+	strm->obuflen = 0;
 	return 0;
 }
 
 static inline int
-lzclose(void *strm)
+lzclose(z_strm *strm)
 {
-	int ret = close(((z_strm *)strm)->sock);
-	LZ4_freeStream(((z_strm *)strm)->hdl.lz);
-	free(strm);
+	lzflush(strm);
+	free(strm->hdl.z.cbuf);
+	return strm->nextstrm->strmclose(strm->nextstrm);
+}
+
+static inline const char *
+lzerror(z_strm *strm, int rval)
+{
+	return strm->nextstrm->strmerror(strm->nextstrm, rval);
+}
+#endif
+
+#ifdef HAVE_SNAPPY
+/* snappy wrapped socket */
+static inline int snappyflush(z_strm *strm);
+
+static inline ssize_t
+snappywrite(z_strm *strm, const void *buf, size_t sze)
+{
+	/* ensure we have space available */
+	if (strm->obuflen + sze > METRIC_BUFSIZ)
+		if (snappyflush(strm) != 0)
+			return -1;
+
+	/* append metric to buf */
+	memcpy(strm->obuf + strm->obuflen, buf, sze);
+	strm->obuflen += sze;
+
+	return sze;
+}
+
+static inline int
+snappyflush(z_strm *strm)
+{
+	int cret;
+	char cbuf[METRIC_BUFSIZ];
+	size_t cbuflen = sizeof(cbuf);
+	char *cbufp = cbuf;
+	int oret;
+
+	cret = snappy_compress(strm->obuf, strm->obuflen, cbuf, &cbuflen);
+	/* reset the write position, from this point it will always need to
+	 * restart */
+	strm->obuflen = 0;
+
+	if (cret != SNAPPY_OK)
+		return -1;  /* we must reset/free snappy */
+	while (cbuflen > 0) {
+		oret = strm->nextstrm->strmwrite(strm->nextstrm, cbufp, cbuflen);
+		if (oret < 0)
+			return -1;  /* failure is failure */
+		strm->nextstrm->strmflush(strm->nextstrm);
+		/* update counters to possibly retry the remaining bit */
+		cbufp += oret;
+		cbuflen -= oret;
+	}
+
+	return 0;
+}
+
+static inline int
+snappyclose(z_strm *strm)
+{
+	int ret = strm->nextstrm->strmclose(strm->nextstrm);
 	return ret;
 }
+
+static inline const char *
+snappyerror(z_strm *strm, int rval)
+{
+	return strm->nextstrm->strmerror(strm->nextstrm, rval);
+}
 #endif
 
 #ifdef HAVE_SSL
 /* (Open|Libre)SSL wrapped socket */
 static inline ssize_t
-sslwrite(void *strm, const void *buf, size_t sze)
+sslwrite(z_strm *strm, const void *buf, size_t sze)
 {
-	return (ssize_t)SSL_write((SSL *)strm, buf, (int)sze);
+	return (ssize_t)SSL_write(strm->hdl.ssl, buf, (int)sze);
 }
 
 static inline int
-sslflush(void *strm)
+sslflush(z_strm *strm)
 {
 	/* noop */
+	(void)strm;
 	return 0;
 }
 
 static inline int
-sslclose(void *strm)
+sslclose(z_strm *strm)
 {
-	int sock = SSL_get_fd((SSL *)strm);
-	SSL_free((SSL *)strm);
+	int sock = SSL_get_fd(strm->hdl.ssl);
+	SSL_free(strm->hdl.ssl);
 	return close(sock);
 }
 
+static char _sslerror_buf[256];
 static inline const char *
-sslerror(void *strm, int rval)
+sslerror(z_strm *strm, int rval)
 {
-	int err = SSL_get_error((SSL *)strm, rval);
-	return ERR_reason_error_string(err);
+	int err = SSL_get_error(strm->hdl.ssl, rval);
+	switch (err) {
+		case SSL_ERROR_NONE:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: SSL_ERROR_NONE", err);
+			break;
+		case SSL_ERROR_ZERO_RETURN:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: TLS/SSL connection has been closed", err);
+			break;
+		case SSL_ERROR_WANT_READ:
+		case SSL_ERROR_WANT_WRITE:
+		case SSL_ERROR_WANT_CONNECT:
+		case SSL_ERROR_WANT_ACCEPT:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: the read or write operation did not complete", err);
+			break;
+		case SSL_ERROR_WANT_X509_LOOKUP:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: call callback via SSL_CTX_set_client_cert_cb()", err);
+			break;
+#ifdef SSL_ERROR_WANT_ASYNC
+		case SSL_ERROR_WANT_ASYNC:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: asynchronous engine is still processing data", err);
+			break;
+#endif
+#ifdef SSL_ERROR_WANT_ASYNC_JOB
+		case SSL_ERROR_WANT_ASYNC_JOB:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: no async jobs available in the pool", err);
+			break;
+#endif
+		case SSL_ERROR_SYSCALL:
+			snprintf(_sslerror_buf, sizeof(_sslerror_buf),
+					"%d: I/O error: %s", err, strerror(errno));
+			break;
+		case SSL_ERROR_SSL:
+			ERR_error_string_n(ERR_get_error(),
+					_sslerror_buf, sizeof(_sslerror_buf));
+			break;
+	}
+	return _sslerror_buf;
 }
 #endif
 
@@ -253,11 +502,13 @@ server_queuereader(void *d)
 
 	*metric = NULL;
 
-#define FAIL_WAIT_TIME   6  /* 6 * 250ms = 1.5s */
-#define DISCONNECT_WAIT_TIME   12  /* 12 * 250ms = 3s */
-#define LEN_CRITICAL(Q)  (queue_free(Q) < self->bsize)
 	self->running = 1;
 	while (1) {
+		/* close connection when we're asked to reopen */
+		if (__sync_bool_compare_and_swap(&(self->reopen_con), 1, 0)) {
+			self->strm->strmclose(self->strm);
+			self->fd = -1;
+		}
 		if (queue_len(self->queue) == 0) {
 			/* if we're idling, close the TCP connection, this allows us
 			 * to reduce connections, while keeping the connection alive
@@ -266,14 +517,14 @@ server_queuereader(void *d)
 			if (self->ctype == CON_TCP && self->fd >= 0 &&
 					idle++ > DISCONNECT_WAIT_TIME)
 			{
-				self->strmclose(self->strm);
+				self->strm->strmclose(self->strm);
 				self->fd = -1;
 			}
 			if (idle == 1)
 				/* ensure blocks are pushed out as soon as we're idling,
 				 * this allows compressors to benefit from a larger
-				 * stream of data to gain better compresion */
-				self->strmflush(self->strm);
+				 * stream of data to gain better compression */
+				self->strm->strmflush(self->strm);
 			gettimeofday(&stop, NULL);
 			__sync_add_and_fetch(&(self->ticks), timediff(start, stop));
 			if (__sync_bool_compare_and_swap(&(self->keep_running), 0, 0))
@@ -366,7 +617,8 @@ server_queuereader(void *d)
 				__sync_add_and_fetch(&(self->dropped), 1);
 			}
 			gettimeofday(&stop, NULL);
-			self->ticks += timediff(start, stop);
+			__sync_add_and_fetch(&(self->ticks),
+					timediff(start, stop));
 			if (squeue == NULL) {
 				/* we couldn't do anything, take it easy for a bit */
 				if (__sync_add_and_fetch(&(self->failure), 0) > 1) {
@@ -405,7 +657,8 @@ server_queuereader(void *d)
 				/* re-lookup the address info, if it fails, stay with
 				 * whatever we have such that resolution errors incurred
 				 * after starting the relay won't make it fail */
-				freeaddrinfo(self->saddr);
+				if (self->saddr)
+					freeaddrinfo(self->saddr);
 				snprintf(sport, sizeof(sport), "%u", self->port);
 				if (getaddrinfo(self->ip, sport, self->hint, &saddr) == 0) {
 					self->saddr = saddr;
@@ -616,59 +869,119 @@ server_queuereader(void *d)
 #endif
 
 #ifdef HAVE_GZIP
-			if (self->transport == W_GZIP) {
-				self->strm = gzdopen(self->fd, "w");
-				if (self->strm == Z_NULL) {
-					logerr("failed to open gzip stream: %s\n", strerror(errno));
+			if ((self->transport & 0xFFFF) == W_GZIP) {
+				self->strm->hdl.gz = malloc(sizeof(z_stream));
+				if (self->strm->hdl.gz != NULL) {
+					self->strm->hdl.gz->zalloc = Z_NULL;
+					self->strm->hdl.gz->zfree = Z_NULL;
+					self->strm->hdl.gz->opaque = Z_NULL;
+					self->strm->hdl.gz->next_in = Z_NULL;
+				}
+				if (self->strm->hdl.gz == NULL ||
+						deflateInit2(self->strm->hdl.gz,
+							Z_DEFAULT_COMPRESSION,
+							Z_DEFLATED,
+							15 + 16,
+							8,
+							Z_DEFAULT_STRATEGY) != Z_OK)
+				{
+					logerr("failed to open gzip stream: out of memory\n");
 					close(self->fd);
 					self->fd = -1;
 					continue;
 				}
+				self->strm->obuflen = 0;
 			}
 #endif
 #ifdef HAVE_LZ4
-			if (self->transport == W_LZ4) {
-				z_strm *s = (z_strm *)malloc(sizeof(z_strm));
-				if (s == NULL) {
-					logerr("failed to allocate lz4 stream: %s\n",
-							strerror(errno));
+			if ((self->transport & 0xFFFF) == W_LZ4) {
+				self->strm->obuflen = 0;
+
+				/* get the maximum size that should ever be required and allocate for it */
+
+				self->strm->hdl.z.cbuflen = LZ4F_compressFrameBound(sizeof(self->strm->obuf), NULL);
+				if ((self->strm->hdl.z.cbuf = malloc(self->strm->hdl.z.cbuflen)) == NULL) {
+					logerr("Failed to allocate %lu bytes for compressed LZ4 data\n", self->strm->hdl.z.cbuflen);
 					close(self->fd);
 					self->fd = -1;
 					continue;
 				}
-				s->sock = self->fd;
-				s->hdl.lz = LZ4_createStream();
-				if (s->hdl.lz == NULL) {
-					logerr("failed to create lz4 stream: out of memory\n");
-					free(s);
-					close(self->fd);
-					self->fd = -1;
-					continue;
-				}
-				self->strm = s;
 			}
 #endif
+#ifdef HAVE_SNAPPY
+			if ((self->transport & 0xFFFF) == W_SNAPPY) {
+				self->strm->obuflen = 0;
+			}
+#endif
+
 #ifdef HAVE_SSL
-			if (self->transport == W_SSL) {
+			if (self->transport & W_SSL) {
 				int rv;
-				self->strm = SSL_new(self->ctx);
-				if (SSL_set_fd(self->strm, self->fd) == 0) {
+				z_strm *sstrm;
+
+				if ((self->transport & 0xFFFF) == W_PLAIN) {
+ 					/* just SSL, nothing else */
+					sstrm = self->strm;
+				} else {
+					sstrm = self->strm->nextstrm;
+				}
+
+				sstrm->hdl.ssl = SSL_new(sstrm->ctx);
+				SSL_set_tlsext_host_name(sstrm->hdl.ssl, self->ip);
+				if (SSL_set_fd(sstrm->hdl.ssl, self->fd) == 0) {
 					logerr("failed to SSL_set_fd: %s\n",
 							ERR_reason_error_string(ERR_get_error()));
-					self->strmclose(self->strm);
+					sstrm->strmclose(sstrm);
 					self->fd = -1;
 					continue;
 				}
-				if ((rv = SSL_connect(self->strm)) != 1) {
-					logerr("failed to  connect ssl stream: %s\n",
-							ERR_reason_error_string(
-								SSL_get_error(self->strm, rv)));
-					self->strmclose(self->strm);
+				if (self->transport & W_MTLS) {
+					int ret;
+
+					/* issue #444 */
+					ret = SSL_use_certificate_file(sstrm->hdl.ssl,
+												   self->mtlspemcert,
+												   SSL_FILETYPE_PEM);
+					if (ret == 1) {
+						ret = SSL_use_PrivateKey_file(sstrm->hdl.ssl,
+													  self->mtlspemkey,
+													  SSL_FILETYPE_PEM);
+					}
+
+					if (ret != 1) {
+						logerr("failed to enable mTLS: %s\n",
+							   sslerror(sstrm, ret));
+						sstrm->strmclose(sstrm);
+						self->fd = -1;
+						continue;
+					}
+				}
+				if ((rv = SSL_connect(sstrm->hdl.ssl)) != 1) {
+					logerr("failed to connect ssl stream: %s\n",
+							sslerror(sstrm, rv));
+					sstrm->strmclose(sstrm);
 					self->fd = -1;
 					continue;
 				}
-			}
+				if ((rv = SSL_get_verify_result(sstrm->hdl.ssl)) != X509_V_OK) {
+					logerr("failed to verify ssl certificate: %s\n",
+							sslerror(sstrm, rv));
+					sstrm->strmclose(sstrm);
+					self->fd = -1;
+					continue;
+				}
+			} else
 #endif
+			{
+				z_strm *pstrm;
+
+				if (self->transport == W_PLAIN) { /* just plain socket */
+					pstrm = self->strm;
+				} else {
+					pstrm = self->strm->nextstrm;
+				}
+				pstrm->hdl.sock = self->fd;
+			}
 		}
 
 		/* send up to batch size */
@@ -697,14 +1010,14 @@ server_queuereader(void *d)
 		}
 
 		for (; *metric != NULL; metric++) {
-			len = strlen(*metric);
+			len = *(size_t *)(*metric);
 			/* Write to the stream, this may not succeed completely due
 			 * to flow control and whatnot, which the docs suggest need
 			 * resuming to complete.  So, use a loop, but to avoid
 			 * getting endlessly stuck on this, only try a limited
 			 * number of times for a single metric. */
-			for (cnt = 0, p = *metric; cnt < 10; cnt++) {
-				if ((slen = self->strmwrite(self->strm, p, len)) != len) {
+			for (cnt = 0, p = *metric + sizeof(size_t); cnt < 10; cnt++) {
+				if ((slen = self->strm->strmwrite(self->strm, p, len)) != len) {
 					if (slen >= 0) {
 						p += slen;
 						len -= slen;
@@ -725,16 +1038,18 @@ server_queuereader(void *d)
 						__sync_fetch_and_add(&(self->failure), 1) == 0)
 					logerr("failed to write() to %s:%u: %s\n",
 							self->ip, self->port,
-							(slen < 0 ? self->strmerror(self->strm, slen) :
+							(slen < 0 ?
+							 self->strm->strmerror(self->strm, slen) :
 							 "incomplete write"));
-				self->strmclose(self->strm);
+				self->strm->strmclose(self->strm);
 				self->fd = -1;
 				/* put back stuff we couldn't process */
 				for (; *metric != NULL; metric++) {
 					if (!queue_putback(self->queue, *metric)) {
 						if (mode & MODE_DEBUG)
 							logerr("server %s:%u: dropping metric: %s",
-									self->ip, self->port, *metric);
+									self->ip, self->port,
+									*metric + sizeof(size_t));
 						free((char *)*metric);
 						__sync_add_and_fetch(&(self->dropped), 1);
 					}
@@ -757,7 +1072,7 @@ server_queuereader(void *d)
 	__sync_and_and_fetch(&(self->running), 0);
 
 	if (self->fd >= 0)
-		self->strmclose(self->strm);
+		self->strm->strmclose(self->strm);
 	if (secpos != NULL)
 		free(secpos);
 	return NULL;
@@ -774,6 +1089,8 @@ server_new(
 		unsigned short port,
 		con_type type,
 		con_trnsp transport,
+		char *mtlspemcert,
+		char *mtlspemkey,
 		con_proto ctype,
 		struct addrinfo *saddr,
 		struct addrinfo *hint,
@@ -790,6 +1107,8 @@ server_new(
 
 	ret->type = type;
 	ret->transport = transport;
+	ret->mtlspemcert = mtlspemcert == NULL ? NULL : strdup(mtlspemcert);
+	ret->mtlspemkey = mtlspemkey == NULL ? NULL : strdup(mtlspemkey);
 	ret->ctype = ctype;
 	ret->tid = 0;
 	ret->secondaries = NULL;
@@ -811,54 +1130,111 @@ server_new(
 		return NULL;
 	}
 	ret->fd = -1;
-	if (transport == W_PLAIN) {
-		ret->strm = &(ret->fd);
-		ret->strmwrite = &sockwrite;
-		ret->strmflush = &sockflush;
-		ret->strmclose = &sockclose;
-		ret->strmerror = &sockerror;
+	if ((ret->strm = malloc(sizeof(z_strm))) == NULL) {
+		free((char *)ret->ip);
+		free(ret->batch);
+		free(ret);
+		return NULL;
+	}
+	ret->strm->nextstrm = NULL;
+
+	/* setup normal or SSL-wrapped socket first */
+#ifdef HAVE_SSL
+	if (transport & W_SSL) {
+		/* create an auto-negotiate context */
+		const SSL_METHOD *m = SSLv23_client_method();
+		ret->strm->ctx = SSL_CTX_new(m);
+
+		if (sslCA != NULL) {
+			if (ret->strm->ctx == NULL ||
+				SSL_CTX_load_verify_locations(ret->strm->ctx,
+											   sslCAisdir ? NULL : sslCA,
+											   sslCAisdir ? sslCA : NULL) == 0)
+			{
+				char *err = ERR_error_string(ERR_get_error(), NULL);
+				logerr("failed to load SSL verify locations from %s for "
+						"%s:%d: %s\n", sslCA, ret->ip, ret->port, err);
+				free((char *)ret->ip);
+				free(ret->batch);
+				free(ret->strm);
+				free(ret);
+				return NULL;
+			}
+		}
+		SSL_CTX_set_verify(ret->strm->ctx, SSL_VERIFY_PEER, NULL);
+
+		ret->strm->strmwrite = &sslwrite;
+		ret->strm->strmflush = &sslflush;
+		ret->strm->strmclose = &sslclose;
+		ret->strm->strmerror = &sslerror;
+	} else
+#endif
+	{
+		ret->strm->strmwrite = &sockwrite;
+		ret->strm->strmflush = &sockflush;
+		ret->strm->strmclose = &sockclose;
+		ret->strm->strmerror = &sockerror;
+	}
+	
+	/* now see if we have a compressor defined */
+	if ((transport & 0xFFFF) == W_PLAIN) {
+		/* catch noop */
 	}
 #ifdef HAVE_GZIP
-	else if (transport == W_GZIP) {
-		ret->strmwrite = &gzipwrite;
-		ret->strmflush = &gzipflush;
-		ret->strmclose = &gzipclose;
-		ret->strmerror = &sockerror;
-	}
-#endif
-#ifdef HAVE_LZ4
-	else if (transport == W_LZ4) {
-		ret->strmwrite = &lzwrite;
-		ret->strmflush = &lzflush;
-		ret->strmclose = &lzclose;
-		ret->strmerror = &sockerror;
-	}
-#endif
-#ifdef HAVE_SSL
-	else if (transport == W_SSL) {
-		/* create a auto-negotiate context */
-		const SSL_METHOD *m = SSLv23_client_method();
-		ret->ctx = SSL_CTX_new(m);
-		
-		if (ret->ctx == NULL) {
-			char *err = ERR_error_string(ERR_get_error(), NULL);
-			logerr("failed to create SSL context for server "
-					"%s:%d: %s\n", ret->ip, ret->port, err);
+	else if ((transport & 0xFFFF) == W_GZIP) {
+		z_strm *gzstrm = malloc(sizeof(z_strm));
+		if (gzstrm == NULL) {
 			free((char *)ret->ip);
+			free(ret->batch);
+			free(ret->strm);
 			free(ret);
 			return NULL;
 		}
-
-		ret->strmwrite = &sslwrite;
-		ret->strmflush = &sslflush;
-		ret->strmclose = &sslclose;
-		ret->strmerror = &sslerror;
+		gzstrm->strmwrite = &gzipwrite;
+		gzstrm->strmflush = &gzipflush;
+		gzstrm->strmclose = &gzipclose;
+		gzstrm->strmerror = &gziperror;
+		gzstrm->nextstrm = ret->strm;
+		ret->strm = gzstrm;
 	}
 #endif
-	else {
-		logerr("no transport type defined for server!!! (this is a BUG)\n");
-		assert(0);
+#ifdef HAVE_LZ4
+	else if ((transport & 0xFFFF) == W_LZ4) {
+		z_strm *lzstrm = malloc(sizeof(z_strm));
+		if (lzstrm == NULL) {
+			free((char *)ret->ip);
+			free(ret->batch);
+			free(ret->strm);
+			free(ret);
+			return NULL;
+		}
+		lzstrm->strmwrite = &lzwrite;
+		lzstrm->strmflush = &lzflush;
+		lzstrm->strmclose = &lzclose;
+		lzstrm->strmerror = &lzerror;
+		lzstrm->nextstrm = ret->strm;
+		ret->strm = lzstrm;
 	}
+#endif
+#ifdef HAVE_SNAPPY
+	else if ((transport & 0xFFFF) == W_SNAPPY) {
+		z_strm *snpstrm = malloc(sizeof(z_strm));
+		if (snpstrm == NULL) {
+			free((char *)ret->ip);
+			free(ret->batch);
+			free(ret->strm);
+			free(ret);
+			return NULL;
+		}
+		snpstrm->strmwrite = &snappywrite;
+		snpstrm->strmflush = &snappyflush;
+		snpstrm->strmclose = &snappyclose;
+		snpstrm->strmerror = &snappyerror;
+		snpstrm->nextstrm = ret->strm;
+		ret->strm = snpstrm;
+	}
+#endif
+
 	ret->saddr = saddr;
 	ret->reresolve = 0;
 	ret->hint = NULL;
@@ -868,8 +1244,6 @@ server_new(
 	}
 	ret->queue = queue_new(qsize);
 	if (ret->queue == NULL) {
-		if (ret->hint)
-			free(ret->hint);
 		free(ret->batch);
 		free((char *)ret->ip);
 		free(ret);
@@ -1018,6 +1392,15 @@ server_send(server *s, const char *d, char force)
 	return 1;
 }
 
+/**
+ * Flag worker main loop to close the connection if currently open.
+ */
+void
+server_closecon(server *s)
+{
+	__sync_bool_compare_and_swap(&(s->reopen_con), 0, 1);
+}
+
 /**
  * Tells this server to finish sending pending items from its queue.
  */
@@ -1091,6 +1474,10 @@ server_free(server *s) {
 	}
 
 	queue_destroy(s->queue);
+#ifdef HAVE_SSL
+	if (s->transport & W_SSL)
+		SSL_CTX_free(s->strm->ctx);
+#endif
 	free(s->batch);
 	if (s->instance)
 		free(s->instance);
@@ -1098,7 +1485,14 @@ server_free(server *s) {
 		freeaddrinfo(s->saddr);
 	if (s->hint)
 		free(s->hint);
+	if (s->mtlspemcert)
+		free(s->mtlspemcert);
+	if (s->mtlspemkey)
+		free(s->mtlspemkey);
 	free((char *)s->ip);
+	if (s->strm->nextstrm != NULL)
+		free(s->strm->nextstrm);
+	free(s->strm);
 	s->ip = NULL;
 	free(s);
 }
@@ -1202,7 +1596,10 @@ server_failed(server *s)
 {
 	if (s == NULL)
 		return 0;
-	return __sync_add_and_fetch(&(s->failure), 0);
+	if (!s->failover && s->secondariescnt > 0)
+		return __sync_add_and_fetch(&(s->failure), 0) >= FAIL_WAIT_TIME;
+	else
+		return __sync_add_and_fetch(&(s->failure), 0) > 0;
 }
 
 /**
